package gl

/*
#include "gl.h"
*/
import "C"

import (
	"fmt"
	"unsafe"
)

func GetBufferParameteriv(target, value Enum) []int {
	var param [4]int
	C.glGetBufferParameteriv(C.GLenum(target), C.GLenum(value), (*C.GLint)(unsafe.Pointer(&param)))
	return param[:]
}

func GetBufferParameteri(target, value Enum) int {
	param := GetBufferParameteriv(target, value)
	return param[0]
}

func GenBuffers(n int) []Buffer {
	buffers := make([]Buffer, n)
	C.glGenBuffers(C.GLsizei(n), (*C.GLuint)(&buffers[0]))
	return buffers
}

func DeleteBuffers(buffers []Buffer) {
	C.glDeleteBuffers(C.GLsizei(len(buffers)), (*C.GLuint)(&buffers[0]))
}

func DeleteVertexArrays(vao []VAO) {
	C.glDeleteVertexArrays(C.GLsizei(len(vao)), (*C.GLuint)(&vao[0]))
}

func BindBuffer(target Enum, buffer Buffer) {
	C.glBindBuffer(C.GLenum(target), C.GLuint(buffer))
}

func BufferData(target Enum, size int, data interface{}, usage Enum) {
	p, datasize := getptr(data)
	if size < datasize {
		panic(fmt.Sprintf("gl: buffer data size too small: %v %v", size, datasize))
	}
	C.glBufferData(C.GLenum(target), C.GLsizeiptr(size), p, C.GLenum(usage))
}

func BufferSubData(target Enum, offset int, data interface{}) {
	p, size := getptr(data)
	C.glBufferSubData(C.GLenum(target), C.GLintptr(offset), C.GLsizeiptr(size), p)
}

func MapBufferRange(target Enum, offset, length int, access Bitfield) unsafe.Pointer {
	return unsafe.Pointer(C.glMapBufferRange(C.GLenum(target), C.GLintptr(offset), C.GLsizeiptr(length), C.GLbitfield(access)))
}

func NewVBO(target Enum, size int, data interface{}, usage Enum) VBO {
	defer PanicOnError()
	buf := GenBuffers(1)
	vbo := VBO(buf[0])
	vbo.Bind()
	BufferData(target, size, data, usage)
	return vbo
}

func (v VBO) Bind() {
	BindBuffer(ARRAY_BUFFER, Buffer(v))
}

func (v VBO) Unbind() {
	BindBuffer(ARRAY_BUFFER, 0)
}

func (v VBO) SubData(offset int, data interface{}) {
	BufferSubData(ARRAY_BUFFER, offset, data)
}

func (v VBO) MapRange(offset, length int, access Bitfield) unsafe.Pointer {
	v.Bind()
	return MapBufferRange(ARRAY_BUFFER, offset, length, access)
}

func (v VBO) Free() {
	DeleteBuffers([]Buffer{Buffer(v)})
}

func GenVertexArrays(n int) []VAO {
	vao := make([]VAO, n)
	C.glGenVertexArrays(C.GLsizei(n), (*C.GLuint)(&vao[0]))
	return vao
}

func EnableVertexAttribArray(index Attribute) {
	C.glEnableVertexAttribArray(C.GLuint(index))
}

func DisableVertexAttribArray(index Attribute) {
	C.glDisableVertexAttribArray(C.GLuint(index))
}

func VertexAttribPointer(index Attribute, size int, typ Enum, normalized Boolean, stride int, data interface{}) {
	p, _ := getptr(data)
	C.glVertexAttribPointer(C.GLuint(index), C.GLint(size), C.GLenum(typ),
		C.GLboolean(normalized), C.GLsizei(stride), p)
}

func NewVAO() VAO {
	vao := GenVertexArrays(1)
	return vao[0]
}

func (v VAO) Bind() {
	C.glBindVertexArray(C.GLuint(v))
}

func (v VAO) Unbind() {
	C.glBindVertexArray(0)
}

func (v VAO) Enable(index ...Attribute) {
	for _, i := range index {
		EnableVertexAttribArray(i)
	}
}

func (v VAO) Disable(index ...Attribute) {
	for _, i := range index {
		DisableVertexAttribArray(i)
	}
}

func (v VAO) Attrib(index Attribute, size int, typ Enum, normalized Boolean, stride int, data interface{}) {
	VertexAttribPointer(index, size, typ, normalized, stride, data)
}

func (v VAO) Free() {
	DeleteVertexArrays([]VAO{v})
}
