package gl

/*
#include "gl.h"
#include <stdlib.h>
*/
import "C"

import (
	"errors"
	"fmt"
	"reflect"
	"unsafe"
)

func Init() error {
	if status := Enum(C.glewInit()); status != GLEW_OK {
		return errors.New(GLEWErrorString(status))
	}
	ClearError()
	return nil
}

func GLEWErrorString(status Enum) string {
	str := (*C.char)((unsafe.Pointer)(C.glewGetErrorString(C.GLenum(status))))
	return C.GoString(str)
}

func Error() Enum {
	return Enum(C.glGetError())
}

func ErrorString(err Enum) string {
	return C.GoString((*C.char)((unsafe.Pointer)(C.gluErrorString((C.GLenum)(err)))))
}

func PanicOnError() {
	ok := true
	for {
		e := Error()
		if e == NO_ERROR {
			break
		}

		fmt.Println("gl error: ", ErrorString(e))
		ok = false
	}

	if !ok {
		panic("encountered gl errors")
	}
}

func ClearError() {
	for {
		if e := Error(); e == NO_ERROR {
			break
		}
	}
}

func Clear(mask Bitfield) {
	C.glClear(C.GLbitfield(mask))
}

func ClearColor(r, g, b, a float32) {
	C.glClearColor(C.GLclampf(r), C.GLclampf(g), C.GLclampf(b), C.GLclampf(a))
}

func GetIntegerv(pname Enum) []int {
	v := make([]C.GLint, 8)
	p := make([]int, 8)
	C.glGetIntegerv(C.GLenum(pname), &v[0])
	for i := range v {
		p[i] = int(v[i])
	}
	return p
}

func GetInteger(pname Enum) int {
	p := GetIntegerv(pname)
	return p[0]
}

func GetString(name Enum) string {
	s := C.glGetString(C.GLenum(name))
	return C.GoString((*C.char)(unsafe.Pointer(s)))
}

func DrawArrays(mode Enum, first, count int) {
	C.glDrawArrays(C.GLenum(mode), C.GLint(first), C.GLsizei(count))
}

func Enable(cap Enum) {
	C.glEnable(C.GLenum(cap))
}

func Disable(cap Enum) {
	C.glDisable(C.GLenum(cap))
}

func BlendFunc(sfactor, dfactor Enum) {
	C.glBlendFunc(C.GLenum(sfactor), C.GLenum(dfactor))
}

func CullFace(mode Enum) {
	C.glCullFace(C.GLenum(mode))
}

func FrontFace(mode Enum) {
	C.glFrontFace(C.GLenum(mode))
}

func PolygonMode(face, mode Enum) {
	C.glPolygonMode(C.GLenum(face), C.GLenum(mode))
}

func Viewport(x, y, w, h int) {
	C.glViewport(C.GLint(x), C.GLint(y), C.GLsizei(w), C.GLsizei(h))
}

func DepthRange(n, f float64) {
	C.glDepthRange(C.GLclampd(n), C.GLclampd(f))
}

func DepthRangef(n, f float32) {
	C.glDepthRange(C.GLclampd(n), C.GLclampd(f))
}

func Finish() {
	C.glFinish()
}

func getptr(data interface{}) (unsafe.Pointer, int) {
	var p unsafe.Pointer
	var size int

	if data == nil {
		return p, 0
	}

	switch typ := reflect.TypeOf(data); typ.Kind() {
	case reflect.Uintptr:
		p = unsafe.Pointer(data.(uintptr))
		size = int(unsafe.Sizeof(uintptr(0)))

	case reflect.Int, reflect.Int8, reflect.Int16, reflect.Int32, reflect.Int64:
		value := reflect.ValueOf(data)
		p = unsafe.Pointer(uintptr(value.Int()))
		size = int(unsafe.Sizeof(uintptr(0)))

	case reflect.Uint, reflect.Uint8, reflect.Uint16, reflect.Uint32, reflect.Uint64:
		value := reflect.ValueOf(data)
		p = unsafe.Pointer(uintptr(value.Uint()))
		size = int(unsafe.Sizeof(uintptr(0)))

	case reflect.Slice, reflect.Array:
		value := reflect.ValueOf(data)
		size = int(typ.Elem().Size()) * value.Len()
		p = unsafe.Pointer(value.Pointer())

	case reflect.Ptr, reflect.Struct:
		value := reflect.ValueOf(data)
		size = int(typ.Elem().Size())
		p = unsafe.Pointer(value.Pointer())

	default:
		panic(fmt.Sprintf("gl: unimplemented ptr type: %v", typ))
	}

	return p, size
}
