package gl

/*
#include "gl.h"
#include <stdlib.h>
*/
import "C"

import (
	"fmt"
	"unsafe"
)

func CreateShader(typ Enum) Shader {
	return Shader(C.glCreateShader(C.GLenum(typ)))
}

func (s Shader) Source(src string) {
	csrc := C.CString(src)
	defer C.free(unsafe.Pointer(csrc))
	C.glShaderSource(C.GLuint(s), 1, (**C.GLchar)(unsafe.Pointer(&csrc)), nil)
}

func (s Shader) Compile() {
	C.glCompileShader(C.GLuint(s))
}

func (s Shader) IV(pname Enum) int {
	var params C.GLint
	C.glGetShaderiv(C.GLuint(s), C.GLenum(pname), &params)
	return int(params)
}

func (s Shader) InfoLog() string {
	var info *C.GLchar

	l := s.IV(INFO_LOG_LENGTH)
	info = (*C.GLchar)(C.malloc(C.size_t(l)))
	defer C.free(unsafe.Pointer(info))

	C.glGetShaderInfoLog(C.GLuint(s), C.GLsizei(l), nil, info)
	return C.GoString((*C.char)(unsafe.Pointer(info)))
}

func (s Shader) Delete() {
	C.glDeleteShader(C.GLuint(s))
}

func CreateProgram() Program {
	return Program(C.glCreateProgram())
}

func (p Program) Attach(s Shader) {
	C.glAttachShader(C.GLuint(p), C.GLuint(s))
}

func (p Program) Detach(s Shader) {
	C.glDetachShader(C.GLuint(p), C.GLuint(s))
}

func (p Program) Link() {
	C.glLinkProgram(C.GLuint(p))
}

func (p Program) Use() {
	C.glUseProgram(C.GLuint(p))
}

func (p Program) Delete() {
	C.glDeleteProgram(C.GLuint(p))
}

func (p Program) IV(pname Enum) int {
	var params C.GLint
	C.glGetProgramiv(C.GLuint(p), C.GLenum(pname), &params)
	return int(params)
}

func (p Program) InfoLog() string {
	var info *C.GLchar

	l := p.IV(INFO_LOG_LENGTH)
	info = (*C.GLchar)(C.malloc(C.size_t(l)))
	defer C.free(unsafe.Pointer(info))

	C.glGetProgramInfoLog(C.GLuint(p), C.GLsizei(l), nil, info)
	return C.GoString((*C.char)(unsafe.Pointer(info)))
}

func (p Program) UniformLocation(name string) (Uniform, error) {
	str := C.CString(name)
	defer C.free(unsafe.Pointer(str))
	loc := C.glGetUniformLocation(C.GLuint(p), (*C.GLchar)(unsafe.Pointer(str)))
	if loc < 0 {
		return Uniform(0), fmt.Errorf("could not find uniform: %v", name)
	}

	return Uniform(loc), nil
}

func (p Program) AttribLocation(name string) (Attribute, error) {
	str := C.CString(name)
	defer C.free(unsafe.Pointer(str))
	loc := C.glGetAttribLocation(C.GLuint(p), (*C.GLchar)(unsafe.Pointer(str)))
	if loc < 0 {
		return Attribute(0), fmt.Errorf("could not find attribute: %v", name)
	}

	return Attribute(loc), nil
}

func (p Program) BindAttribLocation(index uint, name string) {
	str := C.CString(name)
	defer C.free(unsafe.Pointer(str))
	C.glBindAttribLocation(C.GLuint(p), C.GLuint(index), (*C.GLchar)(str))
}

func (u Uniform) Set1i(v0 int) {
	C.glUniform1i(C.GLint(u), C.GLint(v0))
}

func (u Uniform) Set2i(v0, v1 int) {
	C.glUniform2i(C.GLint(u), C.GLint(v0), C.GLint(v1))
}

func (u Uniform) Set3i(v0, v1, v2 int) {
	C.glUniform3i(C.GLint(u), C.GLint(v0), C.GLint(v1), C.GLint(v2))
}

func (u Uniform) Set4i(v0, v1, v2, v3 int) {
	C.glUniform4i(C.GLint(u), C.GLint(v0), C.GLint(v1), C.GLint(v2), C.GLint(v3))
}

func (u Uniform) Set1iv(v []int) {
	C.glUniform1iv(C.GLint(u), C.GLsizei(len(v)), (*C.GLint)(unsafe.Pointer(&v[0])))
}

func (u Uniform) Set1fv(v []float32) {
	C.glUniform1fv(C.GLint(u), C.GLsizei(len(v)), (*C.GLfloat)(unsafe.Pointer(&v[0])))
}

func (u Uniform) Set1f(v0 float32) {
	C.glUniform1f(C.GLint(u), C.GLfloat(v0))
}

func (u Uniform) Set2f(v0, v1 float32) {
	C.glUniform2f(C.GLint(u), C.GLfloat(v0), C.GLfloat(v1))
}

func (u Uniform) Set3f(v0, v1, v2 float32) {
	C.glUniform3f(C.GLint(u), C.GLfloat(v0), C.GLfloat(v1), C.GLfloat(v2))
}

func (u Uniform) Set4f(v0, v1, v2, v3 float32) {
	C.glUniform4f(C.GLint(u), C.GLfloat(v0), C.GLfloat(v1), C.GLfloat(v2), C.GLfloat(v3))
}

func (u Uniform) SetMatrix4fv(count int, transpose Boolean, value []float32) {
	C.glUniformMatrix4fv(C.GLint(u), C.GLsizei(count), C.GLboolean(transpose), (*C.GLfloat)(&value[0]))
}
